"""
Copyright 2022, the CVXPY developers.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Tuple

import numpy as np

import cvxpy as cp
from cvxpy.tests.base_test import BaseTest


class TestKron(BaseTest):
    """
    The Kronecker product of matrices M, N is :

        kron(M, N) = [M[0,0] * N   , ..., M[0, end] * N  ]
                     [M[1,0] * N   , ..., M[1, end] * N  ]
                     ...
                     [M[end, 0] * N, ..., M[end, end] * N]
    """

    @staticmethod
    def make_kron_prob(z_dims: Tuple[int],
                       c_dims: Tuple[int],
                       param: bool,
                       var_left: bool,
                       seed: int):
        """
        Construct random nonnegative matrices (C, L) of shapes
        (c_dims, z_dims) respectively. Define an optimization
        problem with a matrix variable of shape z_dims:

            min sum(Z)
            s.t.  kron(Z, C) >= kron(L, C)   ---   if var_left is True
                  kron(C, Z) >= kron(C, L)   ---   if var_left is False
                  Z >= 0

        Regardless of whether var_left is True or False, the optimal
        solution to that problem is Z = L.

        If param is True, then C is defined as a CVXPY Parameter.
        If param is False, then C is a CVXPY Constant.

        A small remark: the constraint that Z >= 0 is redundant.
        It's there because it's easier to set break points that distinguish
        objective canonicalization and constraint canonicalization
        when there's more than one constraint.
        """
        np.random.seed(seed)
        C_value = np.random.rand(*c_dims).round(decimals=2)
        if param:
            C = cp.Parameter(shape=c_dims)
            C.value = C_value
        else:
            C = cp.Constant(C_value)
        Z = cp.Variable(shape=z_dims)
        L = np.random.rand(*Z.shape).round(decimals=2)
        if var_left:
            constraints = [cp.kron(Z, C) >= cp.kron(L, C), Z >= 0]
            # The cvxcore function get_kronl_mat doesn't work when C is a Parameter.
            # We get around this by having kron be non-dpp, but this comes at
            # the price of eliminating the speed benefit of using Parameter objects.
            # We'll eventually need to extend get_kronl_mat so that it supports
            # Parameters. Until then, I'll make a note that tests here DO PASS
            # with the existing get_kronl_mat implementation if we use the following
            # constraints: [cp.kron(Z - L, C) >= 0, Z >= 0].
        else:
            constraints = [cp.kron(C, Z) >= cp.kron(C, L), Z >= 0]
        obj_expr = cp.sum(Z)
        prob = cp.Problem(cp.Minimize(obj_expr), constraints)
        return Z, C, L, prob


class TestKronRightVar(TestKron):

    C_DIMS = [(1, 1), (2, 1), (1, 2), (2, 2)]

    def test_gen_kronr_param(self):
        z_dims = (2, 2)
        for c_dims in TestKronRightVar.C_DIMS:
            Z, C, L, prob = self.make_kron_prob(z_dims, c_dims, param=True,
                                                var_left=False, seed=0)
            prob.solve(solver=cp.CLARABEL)
            self.assertEqual(prob.status, cp.OPTIMAL)
            con_viols = prob.constraints[0].violation()
            self.assertLessEqual(np.max(con_viols), 1e-4)
            self.assertItemsAlmostEqual(Z.value, L, places=4)

    def test_gen_kronr_const(self):
        z_dims = (2, 2)
        for c_dims in TestKronRightVar.C_DIMS:
            Z, C, L, prob = self.make_kron_prob(z_dims, c_dims, param=False,
                                                var_left=False, seed=0)
            prob.solve(solver=cp.CLARABEL)
            self.assertEqual(prob.status, cp.OPTIMAL)
            con_viols = prob.constraints[0].violation()
            self.assertLessEqual(np.max(con_viols), 1e-4)
            self.assertItemsAlmostEqual(Z.value, L, places=4)


class TestKronLeftVar(TestKron):

    C_DIMS = [(1, 1), (2, 1), (1, 2), (2, 2)]

    def symvar_kronl(self, param):
        # Use a symmetric matrix variable
        X = cp.Variable(shape=(2, 2), symmetric=True)
        b_val = 1.5 * np.ones((1, 1))
        if param:
            b = cp.Parameter(shape=(1, 1))
            b.value = b_val
        else:
            b = cp.Constant(b_val)
        L = np.array([[0.5, 1], [2, 3]])
        U = np.array([[10, 11], [12, 13]])
        kronX = cp.kron(X, b)  # should be equal to X

        objective = cp.Minimize(cp.sum(X.flatten(order='F')))
        constraints = [U >= kronX, kronX >= L]
        prob = cp.Problem(objective, constraints)
        prob.solve()

        self.assertItemsAlmostEqual(X.value, np.array([[0.5, 2], [2, 3]]) / 1.5)
        objective = cp.Maximize(cp.sum(X.flatten(order='F')))
        prob = cp.Problem(objective, constraints)
        prob.solve()
        self.assertItemsAlmostEqual(X.value, np.array([[10, 11], [11, 13]]) / 1.5)
        pass

    def test_symvar_kronl_param(self):
        self.symvar_kronl(param=True)

    def test_symvar_kronl_const(self):
        self.symvar_kronl(param=False)

    def scalar_kronl(self, param):
        y = cp.Variable(shape=(1, 1))
        A_val = np.array([[1., 2.], [3., 4.]])
        L = np.array([[0.5, 1], [2, 3]])
        U = np.array([[10, 11], [12, 13]])
        if param:
            A = cp.Parameter(shape=(2, 2))
            A.value = A_val
        else:
            A = cp.Constant(A_val)
        krony = cp.kron(y, A)  # should be equal to y * A
        constraints = [U >= krony, krony >= L]

        objective = cp.Minimize(y)
        prob = cp.Problem(objective, constraints)
        prob.solve()
        self.assertItemsAlmostEqual(y.value, np.array([[np.max(L / A_val)]]))

        objective = cp.Maximize(y)
        prob = cp.Problem(objective, constraints)
        prob.solve()
        self.assertItemsAlmostEqual(y.value, np.array([[np.min(U / A_val)]]))
        pass

    def test_scalar_kronl_param(self):
        self.scalar_kronl(param=True)

    def test_scalar_kronl_const(self):
        self.scalar_kronl(param=False)

    def test_gen_kronl_param(self):
        z_dims = (2, 2)
        for c_dims in TestKronLeftVar.C_DIMS:
            Z, C, L, prob = self.make_kron_prob(z_dims, c_dims, param=True,
                                                var_left=True, seed=0)
            prob.solve(solver=cp.CLARABEL)
            self.assertEqual(prob.status, cp.OPTIMAL)
            con_viols = prob.constraints[0].violation()
            self.assertLessEqual(np.max(con_viols), 1e-4)
            self.assertItemsAlmostEqual(Z.value, L, places=4)

    def test_gen_kronr_const(self):
        z_dims = (2, 2)
        for c_dims in TestKronLeftVar.C_DIMS:
            Z, C, L, prob = self.make_kron_prob(z_dims, c_dims, param=False,
                                                var_left=True, seed=0)
            prob.solve(solver=cp.CLARABEL)
            self.assertEqual(prob.status, cp.OPTIMAL)
            con_viols = prob.constraints[0].violation()
            self.assertLessEqual(np.max(con_viols), 1e-4)
            self.assertItemsAlmostEqual(Z.value, L, places=4)



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from cvxpy import atoms
from cvxpy.atoms.affine import binary_operators as bin_op
from cvxpy.atoms.affine.diag import diag_vec
from cvxpy.atoms.affine.promote import promote
from cvxpy.atoms.affine.upper_tri import upper_tri
from cvxpy.constraints.psd import PSD
from cvxpy.expressions.constants.parameter import Parameter
from cvxpy.expressions.variable import Variable


# Sublevel sets for quasiconvex atoms.
#
# In the below functions, False is a placeholder for an infeasible
# constraint (one that cannot be represented in a DCP way), and True
# is a placeholder for the absence of a constraint
def dist_ratio_sub(expr, t):
    x = expr.args[0]
    a = expr.a
    b = expr.b

    def sublevel_set():
        if t.value > 1:
            return False
        tsq = t.value**2
        return ((1-tsq)*atoms.sum_squares(x) -
                atoms.matmul(2*(a-tsq*b), x) + atoms.sum_squares(a) -
                tsq*atoms.sum_squares(b)) <= 0
    return [sublevel_set]


def mul_sup(expr, t):
    x, y = expr.args
    if x.is_nonneg() and y.is_nonneg():
        return [x >= t * atoms.inv_pos(y)]
    elif x.is_nonpos() and y.is_nonpos():
        return [-x >= t * atoms.inv_pos(-y)]
    else:
        raise ValueError("Incorrect signs.")


def mul_sub(expr, t):
    x, y = expr.args
    if x.is_nonneg() and y.is_nonpos():
        return [y <= t * atoms.inv_pos(x)]
    elif x.is_nonpos() and y.is_nonneg():
        return [x <= t * atoms.inv_pos(y)]
    else:
        raise ValueError("Incorrect signs.")


def ratio_sup(expr, t):
    x, y = expr.args
    if y.is_nonneg():
        return [x >= t * y]
    elif y.is_nonpos():
        return [x <= t * y]
    else:
        raise ValueError("The denominator's sign must be known.")


def ratio_sub(expr, t):
    x, y = expr.args
    if y.is_nonneg():
        return [x <= t * y]
    elif y.is_nonpos():
        return [x >= t * y]
    else:
        raise ValueError("The denominator's sign must be known.")


def length_sub(expr, t):
    arg = expr.args[0]
    if isinstance(t, Parameter):
        def sublevel_set():
            if t.value < 0:
                return False
            if t.value >= arg.size:
                return True
            return arg[int(atoms.floor(t).value):] == 0
        return [sublevel_set]
    else:
        return [arg[int(atoms.floor(t).value):] == 0]


def sign_sup(expr, t):
    x = expr.args[0]

    def superlevel_set():
        if t.value <= -1:
            return True
        elif t.value <= 1:
            return x >= 0
        else:
            return False
    return [superlevel_set]


def sign_sub(expr, t):
    x = expr.args[0]

    def sublevel_set():
        if t.value >= 1:
            return True
        elif t.value >= -1:
            return x <= 0
        else:
            return False
    return [sublevel_set]


def gen_lambda_max_sub(expr, t):
    return [expr.args[0] == expr.args[0].T,
            expr.args[1] >> 0,
            (t * expr.args[1] - expr.args[0] >> 0)]


def condition_number_sub(expr, t):
    A = expr.args[0]
    n = A.shape[0]
    u = Variable(pos=True)

    prom_ut = promote(u * t, (n,))
    prom_u = promote(u, (n,))
    tmp_expr1 = A - diag_vec(prom_u)
    tmp_expr2 = diag_vec(prom_ut) - A

    return [upper_tri(A) == upper_tri(A.T),
            PSD(A),
            PSD(tmp_expr1),
            PSD(tmp_expr2)]


SUBLEVEL_SETS = {
    atoms.multiply: mul_sub,
    bin_op.DivExpression: ratio_sub,
    atoms.length: length_sub,
    atoms.sign: sign_sub,
    atoms.dist_ratio: dist_ratio_sub,
    atoms.gen_lambda_max: gen_lambda_max_sub,
    atoms.condition_number: condition_number_sub,
}


SUPERLEVEL_SETS = {
    atoms.multiply: mul_sup,
    bin_op.DivExpression: ratio_sup,
    atoms.sign: sign_sup,
}


def sublevel(expr, t):
    """Return the t-level sublevel set for `expr`.

    Returned as a constraint phi_t(x) <= 0, where phi_t(x) is convex.
    """
    try:
        return SUBLEVEL_SETS[type(expr)](expr, t)
    except KeyError:
        raise RuntimeError(
                f"The {type(expr)} atom is not yet supported in DQCP. Please "
                "file an issue here: https://github.com/cvxpy/cvxpy/issues")


def superlevel(expr, t):
    """Return the t-level superlevel set for `expr`.

    Returned as a constraint phi_t(x) >= 0, where phi_t(x) is concave.
    """
    try:
        return SUPERLEVEL_SETS[type(expr)](expr, t)
    except KeyError:
        raise RuntimeError(
                f"The {type(expr)} atom is not yet supported in DQCP. Please "
                "file an issue here: https://github.com/cvxpy/cvxpy/issues")



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import warnings

import numpy as np

from cvxpy import settings
from cvxpy.constraints import PowCone3D
from cvxpy.expressions.constants import Constant
from cvxpy.expressions.variable import Variable
from cvxpy.utilities.power_tools import gm_constrs, powcone_constrs
from cvxpy.utilities.solver_context import SolverInfo


def power_exact_canon(expr, args, solver_context: SolverInfo | None = None):
    """Canonicalize Power using power cone constraints."""
    x = args[0]
    p = expr.p_used

    if p == 1:
        return x, []

    shape = expr.shape
    ones = Constant(np.ones(shape))
    if p == 0:
        return ones, []

    t = Variable(shape)

    if 0 < p < 1:
        alpha = float(p)
        return t, powcone_constrs(t, [x, ones], alpha)
    elif p > 1:
        alpha = float(1 / p)
        constrs = powcone_constrs(x, [t, ones], alpha)
        if p % 2 != 0:
            constrs += [x >= 0]
        return t, constrs
    elif p < 0:
        alpha = float(p / (p - 1))
        return t, powcone_constrs(ones, [x, t], alpha)
    else:
        raise NotImplementedError("This power is not yet supported.")


def power_approx_canon(expr, args, solver_context: SolverInfo | None = None):
    """Canonicalize PowerApprox using SOC constraints via rational approximation."""
    x = args[0]
    p = expr.p_used
    w = expr.w

    if p == 1:
        return x, []

    shape = expr.shape
    ones = Constant(np.ones(shape))
    if p == 0:
        return ones, []

    t = Variable(shape)
    if 0 < p < 1:
        constrs = gm_constrs(t, [x, ones], w)
    elif p > 1:
        constrs = gm_constrs(x, [t, ones], w)
    elif p < 0:
        constrs = gm_constrs(ones, [x, t], w)
    else:
        raise NotImplementedError("This power is not yet supported.")

    # Warn if the solver supports power cones and the approximation is poor
    solver_supports_powcone = (
        solver_context is not None and PowCone3D in solver_context.solver_supported_constraints
    )
    if solver_supports_powcone:
        approx_error = getattr(expr, "approx_error", 0.0)
        num_soc = len(constrs)
        if (
            approx_error > settings.POWERCONE_APPROX_ERROR_THRESHOLD
            or num_soc > settings.POWERCONE_APPROX_SOC_THRESHOLD
        ):
            warnings.warn(
                f"Power atom with exponent {float(expr._p_orig)} is being approximated "
                f"with rational {p} (error: {approx_error:.2e}) "
                f"using {num_soc} SOC constraints. "
                f"Consider using approx=False to use power cones instead.",
                stacklevel=6,
            )

    return t, constrs



"""
Copyright 2013 CVXPY Developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

from cvxpy.atoms.affine.sum import sum as cvxpy_sum
from cvxpy.atoms.norm import norm
from cvxpy.atoms.sum_squares import sum_squares


def mean(x, axis=None, keepdims=False):
    """
    Returns the mean of x.
    """
    if axis is None:
        return cvxpy_sum(x, axis, keepdims) / x.size
    elif axis in (0, 1):
        return cvxpy_sum(x, axis, keepdims) / x.shape[axis]
    else:
        raise UserWarning("cp.mean doesn't yet support axis values other than 0 or 1.")


def std(x, axis=None, keepdims=False, ddof=0):
    """
    Returns the standard deviation of x.

    `ddof` is the quantity to use in the Bessel correction.
    """
    if axis is None:
        return norm((x - mean(x)).flatten(order='F'), 2) / np.sqrt(x.size - ddof)
    elif axis in (0, 1):
        return norm(x - mean(x, axis, True), 2, axis=axis, keepdims=keepdims) \
                / np.sqrt(x.shape[axis] - ddof)
    else:
        raise ValueError("cp.std doesn't yet support axis values other than 0 or 1.")

def var(x, axis=None, keepdims=False, ddof=0):
    """
    Returns the variance of x.

    `ddof` is the quantity to use in the Bessel correction.
    """
    if axis is None:
        return sum_squares(x - mean(x)) / (x.size - ddof)
    elif axis in (0, 1):
        raise NotImplementedError(
            """axis and keepdims are not yet supported for var.
            Use square(std(...)) instead.
            """
        )
        # TODO when sum_squares implements axis and keepdims uncomment:
        # return sum_squares(x - mean(x, axis, True), 2, axis=axis, keepdims=keepdims) \
        #         / (x.shape[axis] - ddof)
    else:
        raise ValueError("cp.var doesn't yet support axis values other than 0 or 1.")



from ctypes import c_double, c_int

import numpy as np

import cvxpy.interface as intf
import cvxpy.settings as s
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.qp_solvers.qp_solver import QpSolver
from cvxpy.utilities.citations import CITATION_DICT


class DAQP(QpSolver):
    """QP interface for the DAQP solver.

    This is a simple implementation based on the `official DAQP documentation
    <https://darnstrom.github.io/daqp/>`_.

    .. note::

        DAQP allows for marking individual constraints as "soft", meaning that
        a small user-defined violation is allowed. We currently don't support
        that option. DAQP also supports MIP QPs, which can be compiled by
        CVXPY and could be supported here as a future addition.

    """

    REQUIRES_CONSTR = False
    BOUNDED_VARIABLES = True

    # Map of DAQP exit flags to CVXPY status.
    STATUS_MAP = {1: s.OPTIMAL,
                  2: s.OPTIMAL, # soft optimal, returned if soft constraints
                                # enabled through solver opts (which we don't
                                # currently support)
                  -2: s.SOLVER_ERROR, # Cycling detected (shouldn't happen?)
                  -1: s.INFEASIBLE,
                  -3: s.UNBOUNDED,
                  # TODO to be tested
                  -4: s.USER_LIMIT, # iter limit reached
                  -5: s.SOLVER_ERROR, # non-convex problem, shouldn't happen
                  -6: s.INFEASIBLE} # provided active set (equality
                                    # constraints) infeasible

    # these are solver options that can be passed; we drop any other - good idea?
    ALLOWED_SOLVER_OPTS = (
        'primal_tol',
        'dual_tol',
        'zero_tol',
        'pivot_tol',
        'progress_tol',
        'cycle_tol',
        'iter_limit',
        'fval_bound',
        'eps_prox',
        'eta_prox'
        # ignore the ones about soft constraints and branch-and-bound for now
    )

    def name(self):
        return s.DAQP

    def import_solver(self) -> None:
        import daqp
        daqp

    def invert(self, solution, inverse_data):

        (xstar,fval,exitflag,info) = solution

        attr = {
            s.SOLVE_TIME: info['solve_time'],
            s.SETUP_TIME: info['setup_time'],
        }
        attr[s.EXTRA_STATS] = info

        # Map DAQP statuses back to CVXPY statuses
        status = self.STATUS_MAP.get(exitflag, s.SOLVER_ERROR)

        if status in s.SOLUTION_PRESENT:
            opt_val = fval + inverse_data[s.OFFSET]
            primal_vars = {
                DAQP.VAR_ID:
                intf.DEFAULT_INTF.const_to_matrix(np.array(xstar))
            }
            # Build dual vars dict keyed by constraint IDs
            # DAQP returns duals for [var_bounds; eq_constrs; ineq_constrs]
            # Skip variable bounds (len_primal), then get constraint duals
            len_primal = len(xstar)
            y = np.array(info['lam'][len_primal:])
            n_eq = inverse_data[self.DIMS].zero
            eq_dual = utilities.get_dual_values(
                y[:n_eq],
                utilities.extract_dual_value,
                inverse_data[self.EQ_CONSTR])
            ineq_dual = utilities.get_dual_values(
                y[n_eq:],
                utilities.extract_dual_value,
                inverse_data[self.NEQ_CONSTR])
            dual_vars = {}
            dual_vars.update(eq_dual)
            dual_vars.update(ineq_dual)
            attr[s.NUM_ITERS] = info['iterations']
            sol = Solution(status, opt_val, primal_vars, dual_vars, attr)
        else:
            sol = failure_solution(status, attr)
        return sol

    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts,
                       solver_cache=None):
        """Solve using DAQP.

        .. note::

            We transform from scipy sparse to numpy dense arrays here;
            if the canonicalization code can handle dense canonicalization
            natively it may be better to do so beforehand.

        """
        import daqp

        # using naming conventions in
        # https://darnstrom.github.io/daqp/start/python 

        H = np.array(data[s.P].todense(), dtype=c_double)
        f = np.array(data[s.Q], dtype=c_double)

        # this can probably be made more efficient
        A = np.array(np.concatenate(
                [data[s.A].todense(), data[s.F].todense()]), dtype=c_double)

        # Upper bounds on problem variable.
        if data[s.UPPER_BOUNDS] is None:
            var_upper_bounds = np.ones(len(f), dtype=c_double) * np.inf
        else:
            var_upper_bounds = data[s.UPPER_BOUNDS]
                    
        bupper = np.array(np.concatenate((
                var_upper_bounds,
                data[s.B], 
                data[s.G])),
            dtype=c_double)

        # Lower bounds on problem variable.
        if data[s.LOWER_BOUNDS] is None:
            var_lower_bounds = np.ones(len(f), dtype=c_double) * -np.inf
        else:
            var_lower_bounds = data[s.LOWER_BOUNDS]

        blower = np.array(
            np.concatenate((
                var_lower_bounds,
                data[s.B], 
                -np.inf*np.ones(data[s.G].shape))),
            dtype=c_double)

        sense = np.array(
            np.concatenate((
                # variable bounds, maybe unused
                np.zeros(len(f), dtype=c_int),
                # equality constraints, always active and immutable
                np.ones(len(data[s.B]), dtype=c_int)*5,
                # inequality constraints
                np.zeros(len(data[s.G]), dtype=c_int))),
            dtype=c_int
        )

        # according to https://stackoverflow.com/questions/16266720/find-out-if-a-matrix-is-positive-definite-with-numpy
        # best way to check positive definiteness (since we already know it is symmetric)

        if 'eps_prox' not in solver_opts:
            try:
                np.linalg.cholesky(H)
                is_positive_definite = True
            except np.linalg.LinAlgError:
                is_positive_definite = False
        else:
            is_positive_definite = False # shouldn't be used

        # Overwrite defaults eps_prox
        solver_opts['eps_prox'] = solver_opts.get('eps_prox', 0. if is_positive_definite else 1e-5)
        # This is chosen by benchmarking to a typical problem class
        # https://gist.github.com/enzbus/3d0236c7ed93cff5f0ec3f8587bcd67e
        # Zero (which is default) can't be used in most cases because Cvxpy's
        # canonicalization may add variables that have no quadratic cost.

        used_solver_opts = {
            k:solver_opts[k] for k in solver_opts
            if k in self.ALLOWED_SOLVER_OPTS}

        (xstar,fval,exitflag,info) = daqp.solve(
            H,f,A,bupper,blower,sense,**used_solver_opts)

        return (xstar,fval,exitflag,info)

    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT["DAQP"]


import numpy as np

import cvxpy as cp
from cvxpy import Variable
from cvxpy.atoms.affine.kron import kron
from cvxpy.constraints import OpRelEntrConeQuad
from cvxpy.utilities.solver_context import SolverInfo


def quantum_rel_entr_canon(expr, args, solver_context: SolverInfo | None = None):
    X, Y = args
    n = X.shape[0]
    Imat = np.eye(n)
    e = Imat.ravel().reshape(n**2, 1)
    # assert X.is_symmetric()
    # assert Y.is_symmetric()
    first_arg = cp.atoms.affine.wraps.symmetric_wrap(kron(X, Imat))
    second_arg = cp.atoms.affine.wraps.symmetric_wrap(kron(Imat, Y))
    epi = Variable(shape=first_arg.shape, symmetric=True)
    orec_con = OpRelEntrConeQuad(
        first_arg, second_arg, epi, expr.quad_approx[0], expr.quad_approx[1]
    )
    # at this point we can be certain that we're dealing with an OpRelEnterConeQuad
    # constraint with real inputs. Canonicalize it, and return the results!
    main_con, aux_cons = cp.reductions.cone2cone.approximations.OpRelEntrConeQuad_canon(
        orec_con, None
    )
    constrs = [main_con] + aux_cons
    return e.T @ epi @ e, constrs



"""
Copyright 2025, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Tree node types for balanced binary tree decomposition of cones.
Used by SOCDim3 and Exotic2Common reductions.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass(frozen=True)
class LeafNode:
    """Leaf node: a single 3D cone with original variable indices."""
    cone_id: int
    var_indices: tuple[int, ...]


@dataclass(frozen=True)
class SplitNode:
    """Internal node combining two subtrees with a 3D cone."""
    cone_id: int
    left: TreeNode
    right: TreeNode


@dataclass(frozen=True)
class SingleVarNode:
    """A single variable with no associated cone (n=1 base case in tree decomposition)."""
    var_index: int


@dataclass(frozen=True)
class SpecialNode:
    """Node for special cases (e.g., SOC dim-1, dim-2, dim-4)."""
    node_type: str
    cone_ids: tuple[int, ...]
    var_indices: tuple[int, ...]
    metadata: dict[str, Any] = field(default_factory=dict)


TreeNode = LeafNode | SplitNode | SingleVarNode | SpecialNode


def get_all_cone_ids(node: TreeNode) -> set[int]:
    """Get all constraint IDs from a tree."""
    if isinstance(node, LeafNode):
        return {node.cone_id}
    elif isinstance(node, SplitNode):
        ids = {node.cone_id}
        ids.update(get_all_cone_ids(node.left))
        ids.update(get_all_cone_ids(node.right))
        return ids
    elif isinstance(node, SingleVarNode):
        return set()
    elif isinstance(node, SpecialNode):
        return set(node.cone_ids)
    return set()


def get_leaf_nodes(node: TreeNode) -> list[LeafNode]:
    """Get all leaf nodes in left-to-right order."""
    if isinstance(node, LeafNode):
        return [node]
    elif isinstance(node, SplitNode):
        return get_leaf_nodes(node.left) + get_leaf_nodes(node.right)
    elif isinstance(node, SingleVarNode):
        return []
    elif isinstance(node, SpecialNode):
        return []
    return []


def get_root_cone_id(node: TreeNode) -> int | None:
    """Get the constraint ID of the root cone."""
    if isinstance(node, LeafNode):
        return node.cone_id
    elif isinstance(node, SplitNode):
        return node.cone_id
    elif isinstance(node, SpecialNode):
        return node.cone_ids[-1] if node.cone_ids else None
    return None



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

from cvxpy.atoms.affine.binary_operators import multiply
from cvxpy.atoms.affine.reshape import reshape
from cvxpy.atoms.affine.sum import sum as cvxpy_sum
from cvxpy.atoms.affine.transpose import permute_dims
from cvxpy.utilities.einsum_utilities import (
    find_contraction,
    greedy_path,
    optimal_path,
    parse_einsum_input,
)


def einsum(subscripts, *exprs, optimize="greedy"):
    """Evaluates the Einstein summation convention on the given expressions.

    This atom is the CVXPY analog of NumPy's einsum function `numpy.einsum` [1],
    and it maintains the same syntax and semantics.

    The einsum operation is evaluated by contracting pairs of expressions by
    elementwise multiplication and summation. The order in which the contractions
    are performed affects both the memory usage and the FLOP count required.

    The optimize parameter determines whether to contract expressions using the
    optimal or greedy ordering. The cost to compute the optimal path is exponential
    in the number of distinct subscripts, while the cost to compute the greedy path
    is cubic in the number of distinct subscripts. We typically expect the greedy
    search to produce the optimal path for most problems.

    Examples
    --------
    >>> import cvxpy as cp
    >>> A = cp.Variable((3, 4))
    >>> B = cp.Variable((4, 5))
    >>> # Matrix multiplication
    >>> result = cp.einsum('ij,jk->ik', A, B)
    >>>
    >>> # Trace of a matrix
    >>> C = cp.Variable((3, 3))
    >>> trace = cp.einsum('ii->', C)

    References
    ----------
    [1] https://numpy.org/doc/stable/reference/generated/numpy.einsum.html

    Parameters
    ----------
    subscripts : str
        The subscripts for the einsum operation.
    exprs : Expression
        The expressions to contract.
    optimize : {bool, 'greedy', 'optimal'}, optional
        Whether to contract the expressions using the optimal or greedy ordering.
        Defaults to "greedy".

    Returns
    -------
    Expression
        The contracted expression.
    """
    """Note for maintainers:
    Here, einsum is implemented using the CVXPY sum, multiply, permute_dims, and reshape atoms.
    The implementation proceeds as follows:

    1. Parse and validate the subscripts and the shapes and count of the expressions.
        The core logic is:
        ```python
        input_subscripts, output_subscript = subscripts.split("->")
        ```
        The rest is validation.

    2. Reduce duplicate indices in the expressions.
        Duplicated indices in a subscript pattern represent indexing along a diagonal
        of the corresponding dimensions. For example, the subscript pattern 'ii->i'
        extracts the diagonal of a matrix and 'ii->' takes the trace. For the following steps,
        it is necessary that every dimension of a tensor is uniquely indexed (within the tensor)
        or not indexed (elipsis). We reduce duplicated indices by creating a new tensor
        of reduced dimension by taking the diagonal elements.

    3. Contraction.
        A. If only one input, simply perform an axis sum.
        B. Otherwise, we iterate over pairs of tensors and contract them. Contracting 
        two tensors involves (i) reshaping and permuting them to compatible shapes where 
        corresponding indices align, (ii) performing elementwise multiplication, and 
        (iii) summing over the contracted dimensions. The order of the contractions is 
        given by the contraction path.

    4. Permute the final result to match output subscript order.
        After all contractions, we permute the final result to match the output subscript order.
    """
    # 1. Initial parsing
    dummy_operands = [np.empty(expr.shape, dtype=np.dtype([])) for expr in exprs]
    input_subscripts, output_subscript, _ = parse_einsum_input((subscripts, *dummy_operands))

    input_list = input_subscripts.split(",")
    output_set = set(output_subscript)
    dimension_dict = _validate_arguments(input_list, exprs)

    # 2. Reduce duplicate indices
    operands = []
    reduced_inputs = []
    input_sets = []
    for expr, input in zip(exprs, input_list, strict=True):
        operand, inputs = _initial_reduction(expr, input, dimension_dict)
        operands.append(operand)
        reduced_inputs.append(inputs)
        input_sets.append(set(inputs))

    # 3.A. If only one input, simply perform an axis sum
    if len(operands) == 1:
        return _sum_single_operand(operands[0], reduced_inputs[0], output_subscript)

    path = _get_path(input_sets, output_set, dimension_dict, optimize)

    # 3.B Contract tensors
    for contraction_inds in path:
        # Results of contraction go to the last of the list
        _, input_sets, to_remove, all_labels = find_contraction(
            contraction_inds, input_sets, output_set
        )

        new_operand, new_input = _contract_pair(
            [operands[i] for i in contraction_inds],
            [reduced_inputs[i] for i in contraction_inds],
            to_remove,
            all_labels,
            dimension_dict,
        )

        operands = [operands[i] for i in range(len(operands)) if i not in contraction_inds]
        operands.append(new_operand)
        reduced_inputs = [
            reduced_inputs[i] for i in range(len(reduced_inputs)) if i not in contraction_inds
        ]
        reduced_inputs.append(new_input)

    # 4. After all contractions, permute the final result to match output subscript order
    final_operand = operands[0]
    final_input = reduced_inputs[0]

    # Create permutation to match output subscript order
    if final_input != output_subscript:
        perm = [final_input.index(x) for x in output_subscript]
        final_operand = permute_dims(final_operand, axes=perm)

    return final_operand


def _validate_arguments(input_list, exprs):
    """Validate the input arguments for einsum operation.

    This function checks that the number of expressions matches the number of
    input subscript patterns, and that the dimensions of each expression are
    consistent with the subscript patterns and with each other.

    Parameters
    ----------
    input_list : list of str
        List of subscript patterns for each input expression.
    exprs : list of Expression
        List of CVXPY expressions to be contracted.

    Returns
    -------
    dict
        Dictionary mapping subscript characters to their corresponding dimensions.

    Raises
    ------
    ValueError
        If the number of expressions doesn't match the number of input patterns,
        if a subscript pattern doesn't match the shape of its corresponding expression,
        or if the same subscript character has inconsistent dimensions across expressions.
    """
    dimension_dict = {}

    if len(exprs) != len(input_list):
        raise ValueError(
            f"Number of arguments ({len(exprs)}) doesn't match "
            f"number of input patterns ({len(input_list)})"
        )

    for tnum, term in enumerate(input_list):
        sh = exprs[tnum].shape
        if len(sh) != len(term):
            raise ValueError(
                "Einstein sum subscript %s does not contain the "
                "correct number of indices for operand %d." % (input_list[tnum], tnum)
            )
        for cnum, char in enumerate(term):
            dim = sh[cnum]
            if char in dimension_dict:
                if dimension_dict[char] != dim:
                    raise ValueError(
                        "Size of label '%s' for operand %d (%d) "
                        "does not match previous terms (%d)."
                        % (char, tnum, dim, dimension_dict[char])
                    )
            else:
                dimension_dict[char] = dim

    return dimension_dict


def _get_path(input_sets, output_set, dimension_dict, optimize):
    """Get the contraction path for the einsum operation.

    This function determines the order in which pairs of tensors should be
    contracted to minimize computational cost. The path optimization can use
    either a greedy or optimal path.

    Parameters
    ----------
    input_sets : list of set
        List of sets containing the subscript characters for each input tensor.
    output_set : set
        Set of subscript characters that should appear in the final output.
    dimension_dict : dict
        Dictionary mapping subscript characters to their corresponding dimensions.
    optimize : {bool, 'greedy', 'optimal'}
        Optimization strategy for determining the contraction path.
        - True or 'optimal': Use optimal path (exponential cost)
        - False or 'greedy': Use greedy path (cubic cost)

    Returns
    -------
    list of tuple
        List of tuples, where each tuple contains the indices of the tensors
        to contract in each step of the contraction process.

    Raises
    ------
    ValueError
        If optimize has an invalid value.
    """
    if optimize in {True, "optimal"}:
        return optimal_path(input_sets, output_set, dimension_dict, np.iinfo(np.int32).max)
    elif optimize in {False, "greedy"}:
        return greedy_path(input_sets, output_set, dimension_dict, np.iinfo(np.int32).max)
    else:
        raise ValueError("Invalid value for optimize. Must be True, False, 'optimal', or 'greedy'.")


def _initial_reduction(operand, inputs, dimension_dict):
    """Reduce operands with repeated indices by taking diagonal elements.

    This function handles the case where a subscript pattern contains repeated
    indices (e.g., 'ii' or 'ijj'). For repeated indices, the corresponding
    dimensions are reduced by taking diagonal elements.

    Parameters
    ----------
    operand : Expression
        The CVXPY expression to reduce.
    inputs : str
        The subscript pattern for this operand.
    dimension_dict : dict
        Dictionary mapping subscript characters to their corresponding dimensions.

    Returns
    -------
    tuple
        A tuple containing:
        - Expression: The reduced operand with repeated indices eliminated
        - str: The updated subscript pattern with repeated indices removed
    """
    # Find repeated indices
    counts = {}
    for x in inputs:
        if x in counts:
            counts[x] += 1
        else:
            counts[x] = 1

    # If there are no repeated indices, return the operator and inputs
    repeats = [x for x, ct in counts.items() if ct > 1]
    if len(repeats) == 0:
        return operand, inputs

    # For each repeated index, get the diagonal reduction
    for x in repeats:
        to_reduce = [i for i, y in enumerate(inputs) if y == x]
        to_keep = [i for i in range(len(inputs)) if i not in to_reduce]
        perm = to_reduce + to_keep

        idxs = np.diag_indices(n=dimension_dict[x], ndim=len(to_reduce))

        inputs = x + "".join([inputs[i] for i in to_keep])

        operand = permute_dims(operand, axes=perm)
        operand = operand[idxs]

    # Return the reduced operand and inputs
    return operand, inputs


def _sum_single_operand(operand, input_subscript, output_subscript):
    """Sum a single operand along specified axes.

    This function handles the case where there is only one input operand.
    It sums the operand along axes that appear in the input subscript but
    not in the output subscript, and then permutes the remaining dimensions
    to match the output subscript order.

    Parameters
    ----------
    operand : Expression
        The single CVXPY expression to sum.
    input_subscript : str
        The subscript pattern for the input operand.
    output_subscript : str
        The desired subscript pattern for the output.

    Returns
    -------
    Expression
        The summed and permuted expression with shape matching the output subscript.
    """

    if len(output_subscript) < len(input_subscript):
        idxs = [i for i, x in enumerate(input_subscript) if x not in output_subscript]
        operand = cvxpy_sum(operand, axis=tuple(idxs), keepdims=False)
    elif len(output_subscript) == 0:
        operand = cvxpy_sum(operand, axis=None, keepdims=False)

    remaining = [x for x in input_subscript if x in output_subscript]
    perm = [remaining.index(x) for x in output_subscript]
    operand = permute_dims(operand, axes=perm)
    return operand


def _contract_pair(operands, input_lists, to_remove, all_labels, dimension_dict):
    """Contract a pair of operands by elementwise multiplication and summation.

    This function performs the core contraction operation between two tensors.
    It aligns the operands to compatible shapes, performs elementwise multiplication,
    and then sums over the to-be-removed dimensions.

    Parameters
    ----------
    operands : list of Expression
        List containing exactly two CVXPY expressions to contract.
    input_lists : list of str
        List of subscript patterns for each operand.
    to_remove : set
        Set of subscript characters that should be summed out (contracted).
    all_labels : str
        All subscript characters involved in this contraction.
    dimension_dict : dict
        Dictionary mapping subscript characters to their corresponding dimensions.

    Returns
    -------
    tuple
        A tuple containing:
        - Expression: The contracted result
        - str: The subscript pattern for the contracted result
    """
    # Convert all_labels from set to sorted string for deterministic ordering
    if isinstance(all_labels, set):
        all_labels = "".join(sorted(all_labels))

    # Align and permute the operands to compatible shapes
    aligned_operands = []
    for operand, input_list in zip(operands, input_lists, strict=True):
        extra_dims = tuple(1 for x in all_labels if x not in input_list)
        if len(extra_dims) > 0:
            shape = operand.shape + extra_dims
            input_list += "".join([x for x in all_labels if x not in input_list])
            perm = [input_list.index(x) for x in all_labels]
            operand = reshape(operand, shape=shape, order="C")

        perm = [input_list.index(x) for x in all_labels]
        if perm != list(range(len(perm))):
            operand = permute_dims(operand, axes=perm)

        aligned_operands.append(operand)

    # Elementwise multiply the operands
    new_operand = multiply(aligned_operands[0], aligned_operands[1])

    # Sum the operands along the axes to be removed
    if len(to_remove) == len(all_labels):
        new_operand = cvxpy_sum(new_operand, axis=None, keepdims=False)
    elif len(to_remove) > 0:
        axes_to_remove = [i for i, x in enumerate(all_labels) if x in to_remove]

        # Reduce single-element axis tuple to int
        if len(axes_to_remove) == 1:
            axis = axes_to_remove[0]
        else:
            axis = tuple(axes_to_remove)
        new_operand = cvxpy_sum(new_operand, axis=axis, keepdims=False)

    new_input = "".join([x for x in all_labels if x not in to_remove])

    # Return the contracted operand and inputs
    return new_operand, new_input



"""
Copyright 2025, the CVXPY Authors

Licensed under the Apache License, Version 2.0 (the "License");

you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
import numpy as np

import cvxpy.settings as s
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.qp_solvers.qp_solver import QpSolver
from cvxpy.utilities.citations import CITATION_DICT


class MPAX(QpSolver):
    """An interface for the MPAX solver.
    """

    # Solver capabilities.
    BOUNDED_VARIABLES = True

    STATUS_MAP = {
                    1: s.SOLVER_ERROR, # 1: UNSPECIFIED
                    2: s.OPTIMAL,      # 2: OPTIMAL
                    3: s.INFEASIBLE_OR_UNBOUNDED,   # 3: PRIMAL_INFEASIBLE
                    4: s.INFEASIBLE_OR_UNBOUNDED,    # 4: DUAL_INFEASIBLE
                    5: s.USER_LIMIT,   # 5: TIME_LIMIT
                    6: s.USER_LIMIT,   # 6: ITERATION_LIMIT
                    7: s.SOLVER_ERROR, # 7: NUMERICAL_ERROR
                    8: s.SOLVER_ERROR, # 8# INVALID_PROBLEM
                    9: s.SOLVER_ERROR  # 9: OTHER
                }

    def name(self):
        """The name of the solver.
        """
        return 'MPAX'

    def import_solver(self) -> None:
        """Imports the solver.
        """
        import jax  # noqa F401
        import mpax  # noqa F401

    def invert(self, solution, inverse_data):
        """Returns the solution to the original problem given the inverse_data.
        """
        attr = {}
        status = self.STATUS_MAP[int(solution.termination_status)]
        attr[s.NUM_ITERS] = solution.iteration_count

        if status in s.SOLUTION_PRESENT:
            opt_val = float(solution.primal_objective) + inverse_data[s.OFFSET]
            primal_vars = {
                MPAX.VAR_ID: np.array(solution.primal_solution, dtype=float)
            }
            # Build dual vars dict keyed by constraint IDs
            # MPAX returns dual_solution as [eq_duals; ineq_duals]
            y = np.array(solution.dual_solution, dtype=float)
            n_eq = inverse_data[self.DIMS].zero
            eq_dual = utilities.get_dual_values(
                y[:n_eq],
                utilities.extract_dual_value,
                inverse_data[self.EQ_CONSTR])
            ineq_dual = utilities.get_dual_values(
                y[n_eq:],
                utilities.extract_dual_value,
                inverse_data[self.NEQ_CONSTR])
            dual_vars = {}
            dual_vars.update(eq_dual)
            dual_vars.update(ineq_dual)
            return Solution(status, opt_val, primal_vars, dual_vars, attr)
        else:
            return failure_solution(status, attr)


    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts, solver_cache=None):
        """Returns the result of the call to the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        warm_start : Bool
            Whether to warm_start mpax.
        verbose : Bool
            Control the verbosity.
        solver_opts : dict
            mpax-specific solver options.

        Returns
        -------
        The result returned by MPAX.
        """
        import jax
        import mpax

        P = data[s.P]
        c = data['q']

        A = -data[s.A]
        b = -data[s.B]

        G = -data[s.F]
        h = -data[s.G]

        lb = data[s.LOWER_BOUNDS]
        if lb is None:
            lb = np.full_like(c, -np.inf)
        ub = data[s.UPPER_BOUNDS]
        if ub is None:
            ub = np.full_like(c, np.inf)


        if P.nnz != 0:
            model = mpax.create_qp(P, c, A, b, G, h, lb, ub)
        else:
            model = mpax.create_lp(c, A, b, G, h, lb, ub)

        algorithm = solver_opts.pop('algorithm', None)
        if algorithm is None or algorithm == 'raPDHG':
            algorithm = 'raPDHG'
            alg = mpax.raPDHG
        elif algorithm == 'r2HPDHG':
            alg = mpax.r2HPDHG
        else:
            raise ValueError('Invalid MPAX algorithm')


        if warm_start and solver_cache is not None and \
            self.name() in solver_cache:
            solver = alg(warm_start=True, verbose=verbose, **solver_opts)
            jit_optimize = jax.jit(solver.optimize)
            initial_primal_solution = solver_cache[self.name()].primal_solution
            initial_dual_solution = solver_cache[self.name()].dual_solution
            results = jit_optimize(model,
                                   initial_primal_solution=initial_primal_solution,
                                   initial_dual_solution=initial_dual_solution)
        else:
            solver = alg(warm_start=False, verbose=verbose, **solver_opts)
            jit_optimize = jax.jit(solver.optimize)
            results = jit_optimize(model)

        if solver_cache is not None:
            solver_cache[self.name()] = results

        return results


    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT["MPAX"]



"""
Copyright 2023, the CVXPY Authors

Licensed under the Apache License, Version 2.0 (the "License");

you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

import cvxpy.interface as intf
import cvxpy.settings as s
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.qp_solvers.qp_solver import QpSolver
from cvxpy.utilities.citations import CITATION_DICT


class PIQP(QpSolver):
    """QP interface for the PIQP solver"""

    MIP_CAPABLE = False
    BOUNDED_VARIABLES = True

    # Map of PIQP status to CVXPY status.
    STATUS_MAP = {"PIQP_SOLVED": s.OPTIMAL,
                  "PIQP_MAX_ITER_REACHED": s.USER_LIMIT,
                  "PIQP_PRIMAL_INFEASIBLE": s.INFEASIBLE,
                  "PIQP_DUAL_INFEASIBLE": s.UNBOUNDED}

    def name(self):
        return s.PIQP

    def import_solver(self) -> None:
        import piqp
        piqp

    def invert(self, solution, inverse_data):
        attr = {s.SOLVE_TIME: solution.info.run_time}
        attr[s.EXTRA_STATS] = {"solution": solution}

        # Map PIQP statuses back to CVXPY statuses
        status = self.STATUS_MAP.get(solution.info.status.name, s.SOLVER_ERROR)

        if status in s.SOLUTION_PRESENT:
            opt_val = solution.info.primal_obj + inverse_data[s.OFFSET]
            primal_vars = {
                PIQP.VAR_ID:
                intf.DEFAULT_INTF.const_to_matrix(np.array(solution.x))
            }

            # Build dual vars dict keyed by constraint IDs
            # PIQP returns solution.y (eq_duals) and solution.z or solution.z_u (ineq_duals)
            ineq_duals = solution.z if hasattr(solution, 'z') else solution.z_u
            eq_dual = utilities.get_dual_values(
                solution.y,
                utilities.extract_dual_value,
                inverse_data[self.EQ_CONSTR])
            ineq_dual = utilities.get_dual_values(
                ineq_duals,
                utilities.extract_dual_value,
                inverse_data[self.NEQ_CONSTR])
            dual_vars = {}
            dual_vars.update(eq_dual)
            dual_vars.update(ineq_dual)
            attr[s.NUM_ITERS] = solution.info.iter
            sol = Solution(status, opt_val, primal_vars, dual_vars, attr)
        else:
            sol = failure_solution(status, attr)
        return sol

    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts,
                       solver_cache=None):
        import piqp
        old_interface = float(piqp.__version__.split('.')[0]) == 0 \
                        and float(piqp.__version__.split('.')[1]) <= 5

        solver_opts = solver_opts.copy()

        solver_opts['backend'] = solver_opts.get('backend', 'sparse')
        backend = solver_opts['backend']
        del solver_opts['backend']

        if backend not in ['dense', 'sparse']:
            raise ValueError("Wrong input, backend must be either dense or sparse")

        def update_solver_settings(solver):
            for opt in solver_opts.keys():
                try:
                    solver.settings.__setattr__(opt, solver_opts[opt])
                except TypeError as e:
                    raise TypeError(f"PIQP: Incorrect type for setting '{opt}'.") from e
                except AttributeError as e:
                    raise TypeError(f"PIQP: Unrecognized solver setting '{opt}'.") from e
            solver.settings.verbose = verbose

        structure_changed = True
        if warm_start and solver_cache is not None and self.name() in solver_cache:
            structure_changed = False

            solver, old_data, _ = solver_cache[self.name()]
            new_args = {}

            for key, param in [(s.Q, 'c'), (s.B, 'b'), (s.G, 'h' if old_interface else 'h_u')]:
                if any(data[key] != old_data[key]):
                    new_args[param] = data[key]

            if backend == 'sparse' and data[s.P].data.shape != old_data[s.P].data.shape:
                    structure_changed = True
            elif data[s.P].data.shape != old_data[s.P].data.shape or any(
                    data[s.P].data != old_data[s.P].data):
                new_args['P'] = data[s.P] if backend == 'sparse' else data[s.P].toarray()

            if backend == 'sparse' and data[s.A].data.shape != old_data[s.A].data.shape:
                    structure_changed = True
            elif data[s.A].data.shape != old_data[s.A].data.shape or any(
                    data[s.A].data != old_data[s.A].data):
                new_args['A'] = data[s.A] if backend == 'sparse' else data[s.A].toarray()

            if backend == 'sparse' and data[s.F].data.shape != old_data[s.F].data.shape:
                    structure_changed = True
            elif data[s.F].data.shape != old_data[s.F].data.shape or any(
                    data[s.F].data != old_data[s.F].data):
                new_args['G'] = data[s.F] if backend == 'sparse' else data[s.F].toarray()

            if not old_interface:
                x_lb = data.get(s.LOWER_BOUNDS)
                x_ub = data.get(s.UPPER_BOUNDS)
                old_x_lb = old_data.get(s.LOWER_BOUNDS)
                old_x_ub = old_data.get(s.UPPER_BOUNDS)
                if x_lb is not None and (old_x_lb is None or any(x_lb != old_x_lb)):
                    new_args['x_l'] = x_lb
                if x_ub is not None and (old_x_ub is None or any(x_ub != old_x_ub)):
                    new_args['x_u'] = x_ub

            if backend == 'dense' and not isinstance(solver, piqp.DenseSolver):
                structure_changed = True
            if backend == 'sparse' and not isinstance(solver, piqp.SparseSolver):
                structure_changed = True

            update_solver_settings(solver)

            if not structure_changed and new_args:
                solver.update(**new_args)

        if structure_changed:
            if backend == 'dense':
                solver = piqp.DenseSolver()
            else:
                solver = piqp.SparseSolver()

            update_solver_settings(solver)

            if backend == 'dense':
                # Convert sparse to dense matrices
                P = data[s.P].toarray()
                A = data[s.A].toarray()
                F = data[s.F].toarray()
            else:
                P = data[s.P]
                A = data[s.A]
                F = data[s.F]

            q = data[s.Q]
            b = data[s.B]
            g = data[s.G]

            x_lb = data.get(s.LOWER_BOUNDS)
            x_ub = data.get(s.UPPER_BOUNDS)
            if old_interface:
                solver.setup(P=P, c=q, A=A, b=b, G=F, h=g)
            else:
                setup_kwargs = dict(P=P, c=q, A=A, b=b, G=F, h_u=g)
                if x_lb is not None:
                    setup_kwargs['x_l'] = x_lb
                if x_ub is not None:
                    setup_kwargs['x_u'] = x_ub
                solver.setup(**setup_kwargs)

        solver.solve()

        result = solver.result

        if solver_cache is not None:
            solver_cache[self.name()] = (solver, data, result)

        return result

    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT["PIQP"]


"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np

from cvxpy import Constant
from cvxpy.atoms.affine.binary_operators import outer
from cvxpy.atoms.affine.sum import sum
from cvxpy.atoms.affine.vec import vec
from cvxpy.expressions.variable import Variable
from cvxpy.utilities.solver_context import SolverInfo


def dotsort_canon(expr, args, solver_context: SolverInfo | None = None):
    x = args[0]
    w = args[1]

    if isinstance(w, Constant):
        w_unique, w_counts = np.unique(w.value, return_counts=True)
    else:
        w_unique, w_counts = w, np.ones(w.size)  # Can't group by unique elements for parameters

    # minimize    sum(t) + q @ w_counts
    # subject to  x @ w_unique.T <= t + q.T
    #             0 <= t

    t = Variable((x.size, 1))
    q = Variable((1, w_unique.size))

    obj = sum(t) + q @ w_counts
    x_w_unique_outer_product = outer(vec(x, order='F'), vec(w_unique, order='F'))
    constraints = [x_w_unique_outer_product <= t + q, t >= 0]
    return obj, constraints



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

import numbers
from typing import List, Literal, Tuple

import numpy as np

import cvxpy.lin_ops.lin_op as lo
import cvxpy.lin_ops.lin_utils as lu
import cvxpy.settings as s
from cvxpy.atoms.affine.affine_atom import AffAtom
from cvxpy.atoms.affine.hstack import hstack
from cvxpy.constraints.constraint import Constraint
from cvxpy.expressions.expression import DEFAULT_ORDER_DEPRECATION_MSG, Expression
from cvxpy.utilities.shape import size_from_shape
from cvxpy.utilities.warn import warn


class reshape(AffAtom):
    """
    Reshapes the expression.

    Vectorizes the expression then unvectorizes it into the new shape.
    The entries are reshaped and stored in column-major order, also known
    as Fortran order.

    Parameters
    ----------
    expr : Expression
       The expression to reshape
    shape : tuple or int
        The shape to reshape to
    order : F(ortran) or C
    """

    def __init__(
        self,
        expr,
        shape: int | Tuple[int, ...],
        order: Literal["F", "C", None] = None
    ) -> None:
        if isinstance(shape, numbers.Integral):
            shape = (int(shape),)
        if not s.ALLOW_ND_EXPR and len(shape) > 2:
            raise ValueError("Expressions of dimension greater than 2 "
                             "are not supported.")
        if any(d == -1 for d in shape):
            shape = self._infer_shape(shape, expr.size)

        self._shape = tuple(shape)
        if order is None:
            reshape_order_warning = DEFAULT_ORDER_DEPRECATION_MSG.replace("FUNC_NAME", "reshape")
            warn(reshape_order_warning, FutureWarning)
            order = 'F'
        assert order in ['F', 'C']
        self.order = order
        super(reshape, self).__init__(expr)

    @staticmethod
    def _infer_shape(shape: Tuple[int, ...], size: int) -> Tuple[int, ...]:
        assert shape.count(-1) == 1, "Only one dimension can be -1."
        if len(shape) == 1:
            shape = (size,)
        else:
            unspecified_index = shape.index(-1)
            specified_dims = [d for i, d in enumerate(shape) if i != unspecified_index]
            for d in specified_dims:
                assert d >= 0, "Specified dimension must be nonnegative."
            divisor = int(np.prod(specified_dims))
            unspecified, remainder = np.divmod(size, divisor)
            if remainder != 0:
                raise ValueError(
                    f"Cannot reshape expression of size {size} into shape {shape}."
                )
            shape = tuple(int(unspecified) if d == -1 else d for d in shape)
        return shape

    def is_atom_log_log_convex(self) -> bool:
        """Is the atom log-log convex?
        """
        return True

    def is_atom_log_log_concave(self) -> bool:
        """Is the atom log-log concave?
        """
        return True

    @AffAtom.numpy_numeric
    def numeric(self, values):
        """Reshape the value.
        """
        return np.reshape(values[0], self.shape, order=self.order)

    def validate_arguments(self) -> None:
        """Checks that the new shape has the same number of entries as the old.
        """
        old_len = self.args[0].size
        new_len = size_from_shape(self._shape)
        if not old_len == new_len:
            raise ValueError(
                "Invalid reshape dimensions %s." % (self._shape,)
            )

    def shape_from_args(self) -> Tuple[int, ...]:
        """Returns the shape argument.
        """
        return self._shape

    def get_data(self):
        """Returns info needed to reconstruct the expression besides the args.
        """
        return [self._shape, self.order]

    def graph_implementation(
        self, arg_objs, shape: Tuple[int, ...], data=None
    ) -> Tuple[lo.LinOp, List[Constraint]]:
        """Reshape

        Parameters
        ----------
        arg_objs : list
            LinExpr for each argument.
        shape : tuple
            The shape of the resulting expression.
        data :
            Additional data required by the atom.

        Returns
        -------
        tuple
            (LinOp for objective, list of constraints)
        """
        arg = arg_objs[0]
        if data[1] == 'F':
            return (lu.reshape(arg, shape), [])
        else:  # 'C':
            arg = lu.transpose(arg)
            if len(shape) <= 1:
                return (lu.reshape(arg, shape), [])
            else:
                result = lu.reshape(arg, shape[::-1])
                return (lu.transpose(result), [])


def deep_flatten(x):
    # base cases
    if isinstance(x, Expression):
        if len(x.shape) == 1:
            return x
        else:
            return x.flatten(order='F')
    elif isinstance(x, np.ndarray) or isinstance(x, (int, float)):
        x = Expression.cast_to_const(x)
        return x.flatten(order='F')
    # recursion
    if isinstance(x, list):
        y = []
        for x0 in x:
            x1 = deep_flatten(x0)
            y.append(x1)
        y = hstack(y)
        return y
    msg = 'The input to deep_flatten must be an Expression, a NumPy array, an int'\
          + ' or float, or a nested list thereof. Received input of type %s' % type(x)
    raise ValueError(msg)



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cvxpy.atoms.sum_largest import sum_largest


def cvar(x, beta):
    r"""The conditional value at risk (CVaR) of a random variable represented by
    the vector of samples ``x``.

    For a probability level :math:`\beta \in [0,1)`, CVaR is the expected value of
    ``x`` in the worst :math:`(1-\beta)` fraction of cases. Equivalently, it is the
    average of the :math:`(1-\beta)` fraction of largest values in ``x``.

    Parameters
    ----------
    x : Expression or numeric constant
        A vector of samples representing the distribution. Must be one-dimensional.
    beta : float
        The probability level. Must be in the range :math:`[0, 1)`.
        For example, :math:`\beta = 0.95` gives the average of the worst 5% of outcomes.

    Returns
    -------
    Expression
        .. math::

            \frac{1}{(1-\beta)m} \sum\nolimits_{\text{largest } (1-\beta)m} x_i

        where :math:`m` is the length of :math:`x`. When :math:`(1-\beta)m` is not an
        integer, the fractional part is handled via linear interpolation.
    """
    if not 0 <= beta < 1:
        raise ValueError(f"The probability level beta must be in the range [0, 1), got {beta}")

    if len(x.shape) != 1:
        raise ValueError(f"cvar input must be a 1d array, got shape {x.shape}")

    k = (1 - beta) * x.shape[0]
    return sum_largest(x, k) / k
    



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import collections
import itertools
import math

import numpy as np
import numpy.linalg as LA
import pytest

# Tests atoms by calling them with a constant value.
import cvxpy as cp
import cvxpy.interface as intf
from cvxpy.error import SolverError
from cvxpy.expressions.constants import Constant, Parameter
from cvxpy.expressions.variable import Variable
from cvxpy.problems.problem import Problem
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.settings import CLARABEL, CVXOPT, MOSEK, OSQP, ROBUST_KKTSOLVER, SCS

ROBUST_CVXOPT = "robust_cvxopt"
SOLVER_TO_TOL = {SCS: 1e-2,
                 CLARABEL: 1e-7,
                 OSQP: 1e-1}
SOLVERS_TO_TRY = [CLARABEL, SCS, OSQP]
# Test CVXOPT if installed.
if CVXOPT in INSTALLED_SOLVERS:
    SOLVERS_TO_TRY += [CVXOPT, ROBUST_CVXOPT]
    SOLVER_TO_TOL[CVXOPT] = 1e-7
    SOLVER_TO_TOL[ROBUST_CVXOPT] = 1e-7

# Test MOSEK if installed.
if MOSEK in INSTALLED_SOLVERS:
    SOLVERS_TO_TRY.append(MOSEK)
    SOLVER_TO_TOL[MOSEK] = 1e-6

v_np = np.array([-1., 2, -2]).T

# Defined here to be used in KNOWN_SOLVER_ERRORS


def log_sum_exp_axis_0(x): return cp.log_sum_exp(x, axis=0, keepdims=True)  # noqa E371
def log_sum_exp_axis_1(x): return cp.log_sum_exp(x, axis=1)  # noqa E371


# map from solver name to a list of strings for atoms that fail.
#   old example: KNOWN_SOLVER_ERRORS[cp.MOSEK] = ['xexp']
KNOWN_SOLVER_ERRORS = collections.defaultdict(list)

atoms_minimize = [
    (cp.abs, (2, 2), [[[-5, 2], [-3, 1]]],
     Constant([[5, 2], [3, 1]])),
    (lambda x: cp.cumsum(x, axis=1), (2, 2), [[[-5, 2], [-3, 1]]],
     Constant([[-5, 2], [-8, 3]])),
    (lambda x: cp.cumsum(x, axis=0), (2, 2), [[[-5, 2], [-3, 1]]],
     Constant([[-5, -3], [-3, -2]])),
    (lambda x: cp.cummax(x, axis=1), (2, 2), [[[-5, 2], [-3, 1]]],
     Constant([[-5, 2], [-3, 2]])),
    (lambda x: cp.cummax(x, axis=0), (2, 2), [[[-5, 2], [-3, 1]]],
     Constant([[-5, 2], [-3, 1]])),
    (cp.diag, (2,), [[[-5, 2], [-3, 1]]], Constant([-5, 1])),
    (cp.diag, (2, 2), [[-5, 1]], Constant([[-5, 0], [0, 1]])),
    (cp.exp, (2, 2), [[[1, 0], [2, -1]]],
     Constant([[math.e, 1], [math.e**2, 1.0 / math.e]])),
    (lambda x: cp.xexp(cp.pos(x)), (2, 2), [[[1, 3], [2, .5]]],
     Constant([[math.e, 3 * math.e**3], [2 * math.e**2, 0.5 * math.e**.5]])),
    (cp.huber, (2, 2), [[[0.5, -1.5], [4, 0]]],
     Constant([[0.25, 2], [7, 0]])),
    (lambda x: cp.huber(x, 2.5), (2, 2), [[[0.5, -1.5], [4, 0]]],
     Constant([[0.25, 2.25], [13.75, 0]])),
    (cp.inv_pos, (2, 2), [[[1, 2], [3, 4]]],
     Constant([[1, 1.0 / 2], [1.0 / 3, 1.0 / 4]])),
    (lambda x: (x + Constant(0))**-1, (2, 2), [[[1, 2], [3, 4]]],
     Constant([[1, 1.0 / 2], [1.0 / 3, 1.0 / 4]])),
    (cp.kl_div, tuple(), [math.e, 1], Constant([1])),
    (cp.kl_div, tuple(), [math.e, math.e], Constant([0])),
    (cp.kl_div, (2,), [[math.e, 1], 1], Constant([1, 0])),
    (cp.rel_entr, tuple(), [math.e, 1], Constant([math.e])),
    (cp.rel_entr, tuple(), [math.e, math.e], Constant([0])),
    (cp.rel_entr, (2,), [[math.e, 1], 1], Constant([math.e, 0])),
    # kron with variable in the right operand
    (lambda x: cp.kron(np.array([[1, 2], [3, 4]]), x), (4, 4),
     [np.array([[5, 6], [7, 8]])],
     Constant(np.kron(np.array([[1, 2], [3, 4]]), np.array([[5, 6], [7, 8]])))),
    (lambda x: cp.kron(np.array([[1, 2], [3, 4], [5, 6]]), x), (6, 4),
     [np.array([[5, 6], [7, 8]])],
     Constant(np.kron(np.array([[1, 2], [3, 4], [5, 6]]), np.array([[5, 6], [7, 8]])))),
    (lambda x: cp.kron(np.array([[1, 2], [3, 4]]), x), (6, 4),
     [np.array([[5, 6], [7, 8], [9, 10]])],
     Constant(np.kron(np.array([[1, 2], [3, 4]]), np.array([[5, 6], [7, 8], [9, 10]])))),

    # kron with variable in the left operand
    (lambda x: cp.kron(x, np.array([[1, 2], [3, 4]])), (4, 4),
     [np.array([[5, 6], [7, 8]])],
     Constant(np.kron(np.array([[5, 6], [7, 8]]), np.array([[1, 2], [3, 4]])))),
    (lambda x: cp.kron(x, np.array([[1, 2], [3, 4], [5, 6]])), (6, 4),
     [np.array([[5, 6], [7, 8]])],
     Constant(np.kron(np.array([[5, 6], [7, 8]]), np.array([[1, 2], [3, 4], [5, 6]])))),
    (lambda x: cp.kron(x, np.array([[1, 2], [3, 4]])), (6, 4),
     [np.array([[5, 6], [7, 8], [9, 10]])],
     Constant(np.kron(np.array([[5, 6], [7, 8], [9, 10]]), np.array([[1, 2], [3, 4]])))),
    (cp.lambda_max, tuple(), [[[2, 0], [0, 1]]], Constant([2])),
    (cp.lambda_max, tuple(), [[[2, 0, 0], [0, 3, 0], [0, 0, 1]]], Constant([3])),
    (cp.lambda_max, tuple(), [[[5, 7], [7, -3]]], Constant([9.06225775])),
    (lambda x: cp.lambda_sum_largest(x, 2), tuple(),
     [[[1, 2, 3], [2, 4, 5], [3, 5, 6]]], Constant([11.51572947])),
    (cp.log_sum_exp, tuple(), [[[5, 7], [0, -3]]], Constant([7.1277708268])),
    (log_sum_exp_axis_0, (1, 2),
     [[[5, 7, 1], [0, -3, 6]]], Constant([[7.12910890], [6.00259878]])),
    (log_sum_exp_axis_1, (3,),
     [[[5, 7, 1], [0, -3, 6]]], Constant([5.00671535, 7.0000454, 6.0067153])),
    (cp.logistic, (2, 2),
     [
        [[math.log(5), math.log(7)],
         [0, math.log(0.3)]]],
     Constant(
        [[math.log(6), math.log(8)],
         [math.log(2), math.log(1.3)]])),
    (cp.matrix_frac, tuple(), [[1, 2, 3],
                               [[1, 0, 0],
                                [0, 1, 0],
                                [0, 0, 1]]], Constant([14])),
    (cp.matrix_frac, tuple(), [[1, 2, 3],
                               [[67, 78, 90],
                                [78, 94, 108],
                                [90, 108, 127]]], Constant([0.46557377049180271])),
    (cp.matrix_frac, tuple(), [[[1, 2, 3],
                                [4, 5, 6]],
                               [[67, 78, 90],
                                [78, 94, 108],
                                [90, 108, 127]]], Constant([0.768852459016])),
    (cp.maximum, (2,), [[-5, 2], [-3, 1], 0, [-1, 2]], Constant([0, 2])),
    (cp.maximum, (2, 2), [[[-5, 2], [-3, 1]], 0, [[5, 4], [-1, 2]]],
     Constant([[5, 4], [0, 2]])),
    (cp.max, tuple(), [[[-5, 2], [-3, 1]]], Constant([2])),
    (cp.max, tuple(), [[-5, -10]], Constant([-5])),
    (lambda x: cp.max(x, axis=0, keepdims=True), (1, 2),
     [[[-5, 2], [-3, 1]]], Constant([[2], [1]])),
    (lambda x: cp.max(x, axis=1), (2,), [[[-5, 2], [-3, 1]]], Constant([-3, 2])),
    (lambda x: cp.norm(x, 2), tuple(), [v_np], Constant([3])),
    (lambda x: cp.norm(x, "fro"), tuple(), [[[-1, 2], [3, -4]]],
     Constant([5.47722557])),
    (lambda x: cp.norm(x, 1), tuple(), [v_np], Constant([5])),
    (lambda x: cp.norm(x, 1), tuple(), [[[-1, 2], [3, -4]]],
     Constant([7])),
    (lambda x: cp.norm(x, "inf"), tuple(), [v_np], Constant([2])),
    (lambda x: cp.norm(x, "inf"), tuple(), [[[-1, 2], [3, -4]]],
     Constant([6])),
    (lambda x: cp.norm(x, "nuc"), tuple(), [[[2, 0], [0, 1]]], Constant([3])),
    (lambda x: cp.norm(x, "nuc"), tuple(), [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]],
     Constant([23.173260452512931])),
    (lambda x: cp.norm(x, "nuc"), tuple(), [[[3, 4, 5], [6, 7, 8]]],
     Constant([14.618376738088918])),
    (lambda x: cp.sum_largest(cp.abs(x), 3), tuple(), [[1, 2, 3, -4, -5]], Constant([5 + 4 + 3])),
    (lambda x: cp.dotsort(cp.abs(x), np.array([2, 1, 0.1])), tuple(), [[1, 2, 3, -4, -5]],
     Constant([10 + 4 + .3])),
    (lambda x: cp.mixed_norm(x, 1, 1), tuple(), [[[1, 2], [3, 4], [5, 6]]],
     Constant([21])),
    (lambda x: cp.mixed_norm(x, 1, 1), tuple(), [[[1, 2, 3], [4, 5, 6]]],
     Constant([21])),
    # (lambda x: mixed_norm(x, 2, 1), tuple(), [[[3, 1], [4, math.sqrt(3)]]],
    #     Constant([7])),
    (lambda x: cp.mixed_norm(x, 1, 'inf'), tuple(), [[[1, 4], [5, 6]]],
     Constant([10])),

    (cp.pnorm, tuple(), [[1, 2, 3]], Constant([3.7416573867739413])),
    (lambda x: cp.pnorm(x, 1), tuple(), [[1.1, 2, -3]], Constant([6.1])),
    (lambda x: cp.pnorm(x, 2), tuple(), [[1.1, 2, -3]], Constant([3.7696153649941531])),
    (lambda x: cp.pnorm(x, 2, axis=0), (2,),
     [[[1, 2], [3, 4]]], Constant([math.sqrt(5), 5.]).T),
    (lambda x: cp.pnorm(x, 2, axis=1), (2,),
     [[[1, 2], [4, 5]]], Constant([math.sqrt(17), math.sqrt(29)])),
    (lambda x: cp.pnorm(x, 'inf'), tuple(), [[1.1, 2, -3]], Constant([3])),
    (lambda x: cp.pnorm(x, 3), tuple(), [[1.1, 2, -3]], Constant([3.3120161866074733])),
    (lambda x: cp.pnorm(x, 5.6), tuple(), [[1.1, 2, -3]], Constant([3.0548953718931089])),
    (lambda x: cp.pnorm(x, 1.2), tuple(),
     [[[1, 2, 3], [4, 5, 6]]], Constant([15.971021676279573])),

    (cp.pos, tuple(), [8], Constant([8])),
    (cp.pos, (2,), [[-3, 2]], Constant([0, 2])),
    (cp.neg, (2,), [[-3, 3]], Constant([3, 0])),


    (lambda x: cp.power(x, 1), tuple(), [7.45], Constant([7.45])),
    (lambda x: cp.power(x, 2), tuple(), [7.45], Constant([55.502500000000005])),
    (lambda x: cp.power(x, -1), tuple(), [7.45], Constant([0.1342281879194631])),
    (lambda x: cp.power(x, -.7), tuple(), [7.45], Constant([0.24518314363015764])),
    (lambda x: cp.power(x, -1.34), tuple(), [7.45], Constant([0.06781263100321579])),
    (lambda x: cp.power(x, 1.34), tuple(), [7.45], Constant([14.746515290825071])),

    (cp.quad_over_lin, tuple(), [[[-1, 2, -2], [-1, 2, -2]], 2], Constant([2 * 4.5])),
    (cp.quad_over_lin, tuple(), [v_np, 2], Constant([4.5])),
    (lambda x: cp.norm(x, 2), tuple(), [[[2, 0], [0, 1]]], Constant([2])),
    (lambda x: cp.norm(x, 2), tuple(),
     [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]], Constant([22.368559552680377])),
    (lambda x: cp.scalene(x, 2, 3), (2, 2), [[[-5, 2], [-3, 1]]], Constant([[15, 4], [9, 2]])),
    (cp.square, (2, 2), [[[-5, 2], [-3, 1]]], Constant([[25, 4], [9, 1]])),
    (cp.sum, tuple(), [[[-5, 2], [-3, 1]]], Constant([-5])),
    (lambda x: cp.sum(x, axis=0), (2,), [[[-5, 2], [-3, 1]]], Constant([-3, -2])),
    (lambda x: cp.sum(x, axis=1), (2,), [[[-5, 2], [-3, 1]]], Constant([-8, 3])),
    (lambda x: (x + Constant(0))**2, (2, 2), [[[-5, 2], [-3, 1]]], Constant([[25, 4], [9, 1]])),
    (lambda x: cp.sum_largest(x, 3), tuple(), [[1, 2, 3, 4, 5]], Constant([5 + 4 + 3])),
    (lambda x: cp.sum_largest(x, 3), tuple(),
     [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]], Constant([9 + 10 + 11])),
    (cp.sum_squares, tuple(), [[[-1, 2], [3, -4]]], Constant([30])),
    (cp.trace, tuple(), [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]], Constant([3 + 7 + 11])),
    (cp.trace, tuple(), [[[-5, 2], [-3, 1]]], Constant([-5 + 1])),
    (cp.tv, tuple(), [[1, -1, 2]], Constant([5])),
    (cp.tv, tuple(), [[1, -1, 2]], Constant([5])),
    (cp.tv, tuple(), [[[-5, 2], [-3, 1]]], Constant([math.sqrt(53)])),
    (cp.tv, tuple(), [[[-5, 2], [-3, 1]], [[6, 5], [-4, 3]], [[8, 0], [15, 9]]],
     Constant([LA.norm([7, -1, -8, 2, -10, 7])])),
    (cp.tv, tuple(), [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]], Constant([4 * math.sqrt(10)])),
    (cp.upper_tri, (3,), [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]], Constant([6, 9, 10])),
    # # Advanced indexing.
    (lambda x: x[[1, 2], [0, 2]], (2,),
     [[[3, 4, 5], [6, 7, 8], [9, 10, 11]]], Constant([4, 11])),
    (lambda x: x[[1, 2]], (2, 2), [[[3, 4, 5], [6, 7, 8]]], Constant([[4, 5], [7, 8]])),
    (lambda x: x[np.array([[3, 4, 5], [6, 7, 8]]).T % 2 == 0], (2,), [[[3, 4, 5], [6, 7, 8]]],
     Constant([6, 4, 8])),
    (lambda x: x[2:0:-1], (2,), [[3, 4, 5]], Constant([5, 4])),
    (lambda x: x[2::-1], (3,), [[3, 4, 5]], Constant([5, 4, 3])),
    (lambda x: x[3:0:-1], (2,), [[3, 4, 5]], Constant([5, 4])),
    (lambda x: x[3::-1], (3,), [[3, 4, 5]], Constant([5, 4, 3])),

]


atoms_maximize = [
    (cp.entr, (2, 2), [[[1, math.e], [math.e**2, 1.0 / math.e]]],
     Constant([[0, -math.e], [-2 * math.e**2, 1.0 / math.e]])),
    (cp.log_det, tuple(),
     [[[20, 8, 5, 2],
       [8, 16, 2, 4],
       [5, 2, 5, 2],
       [2, 4, 2, 4]]], Constant([7.7424020218157814])),
    (cp.geo_mean, tuple(), [[4, 1]], Constant([2])),
    (cp.geo_mean, tuple(), [[0.01, 7]], Constant([0.2645751311064591])),
    (cp.geo_mean, tuple(), [[63, 7]], Constant([21])),
    (cp.geo_mean, tuple(), [[1, 10]], Constant([math.sqrt(10)])),
    (lambda x: cp.geo_mean(x, [1, 1]), tuple(), [[1, 10]], Constant([math.sqrt(10)])),
    (lambda x: cp.geo_mean(x, [.4, .8, 4.9]), tuple(),
     [[.5, 1.8, 17]], Constant([10.04921378316062])),
    (cp.harmonic_mean, tuple(), [[1, 2, 3]], Constant([1.6363636363636365])),
    (cp.harmonic_mean, tuple(), [[2.5, 2.5, 2.5, 2.5]], Constant([2.5])),
    (cp.harmonic_mean, tuple(), [[1e-8, 1, 2]], Constant([0])),

    (lambda x: cp.diff(x, 0), (3,), [[1, 2, 3]], Constant([1, 2, 3])),
    (cp.diff, (2,), [[1, 2, 3]], Constant([1, 1])),
    (cp.diff, tuple(), [[1.1, 2.3]], Constant([1.2])),
    (lambda x: cp.diff(x, 2), tuple(), [[1, 2, 3]], Constant([0])),
    (cp.diff, (3,), [[2.1, 1, 4.5, -.1]], Constant([-1.1, 3.5, -4.6])),
    (lambda x: cp.diff(x, 2), (2,), [[2.1, 1, 4.5, -.1]], Constant([4.6, -8.1])),
    (lambda x: cp.diff(x, 1, axis=0), (1, 2), [np.array([[-5, -3], [2, 1]])],
     Constant([[7], [4]])),
    (lambda x: cp.diff(x, 1, axis=1), (2, 1), [np.array([[-5, -3], [2, 1]])],
     Constant([[2, -1]])),

    (lambda x: cp.pnorm(x, .5), tuple(), [[1.1, 2, .1]], Constant([7.724231543909264])),
    (lambda x: cp.pnorm(x, -.4), tuple(), [[1.1, 2, .1]], Constant([0.02713620334])),
    (lambda x: cp.pnorm(x, -1), tuple(), [[1.1, 2, .1]], Constant([0.0876494023904])),
    (lambda x: cp.pnorm(x, -2.3), tuple(), [[1.1, 2, .1]], Constant([0.099781528576])),

    (cp.lambda_min, tuple(), [[[2, 0], [0, 1]]], Constant([1])),
    (cp.lambda_min, tuple(), [[[5, 7], [7, -3]]], Constant([-7.06225775])),
    (lambda x: cp.lambda_sum_smallest(x, 2), tuple(),
     [[[1, 2, 3], [2, 4, 5], [3, 5, 6]]], Constant([-0.34481428])),
    (cp.log, (2, 2), [[[1, math.e], [math.e**2, 1.0 / math.e]]], Constant([[0, 1], [2, -1]])),
    (cp.log1p, (2, 2), [[[0, math.e - 1],
                         [math.e**2 - 1, 1.0 / math.e - 1]]], Constant([[0, 1], [2, -1]])),
    (cp.minimum, (2,), [[-5, 2], [-3, 1], 0, [1, 2]], Constant([-5, 0])),
    (cp.minimum, (2, 2), [[[-5, 2], [-3, -1]],
                          0,
                          [[5, 4], [-1, 2]]], Constant([[-5, 0], [-3, -1]])),
    (cp.min, tuple(), [[[-5, 2], [-3, 1]]], Constant([-5])),
    (cp.min, tuple(), [[-5, -10]], Constant([-10])),
    (lambda x: x**0.25, tuple(), [7.45], Constant([7.45**0.25])),
    (lambda x: x**0.32, (2,), [[7.45, 3.9]], Constant(np.power(np.array([7.45, 3.9]), 0.32))),
    (lambda x: x**0.9, (2, 2), [[[7.45, 2.2],
                                 [4, 7]]], Constant(np.power(np.array([[7.45, 2.2],
                                                                       [4, 7]]).T, 0.9))),
    (cp.sqrt, (2, 2), [[[2, 4], [16, 1]]], Constant([[1.414213562373095, 2], [4, 1]])),
    (lambda x: cp.sum_smallest(x, 3), tuple(), [[-1, 2, 3, 4, 5]], Constant([-1 + 2 + 3])),
    (lambda x: cp.sum_smallest(x, 4), tuple(),
     [[[-3, -4, 5], [6, 7, 8], [9, 10, 11]]], Constant([-3 - 4 + 5 + 6])),
    (lambda x: (x + Constant(0))**0.5, (2, 2),
     [[[2, 4], [16, 1]]], Constant([[1.414213562373095, 2], [4, 1]])),
]


def check_solver(prob, solver_name) -> bool:
    """Can the solver solve the problem?
    """
    atom_str = str(prob.objective.args[0])
    for bad_atom_name in KNOWN_SOLVER_ERRORS[solver_name]:
        if bad_atom_name in atom_str:
            return False
    try:
        if solver_name == ROBUST_CVXOPT:
            solver_name = CVXOPT

        prob._construct_chain(solver=solver_name)

        return True
    except SolverError:
        return False
    except Exception:
        raise


# Tests numeric version of atoms.
def run_atom(atom, problem, obj_val, solver, verbose: bool = False) -> None:
    assert problem.is_dcp()
    print(problem)
    if verbose:
        print(problem.objective)
        print(problem.constraints)
        print("solver", solver)
    if check_solver(problem, solver):
        tolerance = SOLVER_TO_TOL[solver]

        try:
            if solver == ROBUST_CVXOPT:
                result = problem.solve(solver=CVXOPT, verbose=verbose,
                                       kktsolver=ROBUST_KKTSOLVER)
            else:
                result = problem.solve(solver=solver, verbose=verbose)
        except SolverError as e:
            if (atom, solver) in KNOWN_SOLVER_ERRORS:
                return
            raise e

        if verbose:
            print(result)
            print(obj_val)
        assert (-tolerance <= (result - obj_val) / (1 + np.abs(obj_val)) <= tolerance)


def get_indices(size):
    """Get indices for dimension.
    """
    if len(size) == 0:
        return [0]
    elif len(size) == 1:
        return range(size[0])
    else:
        return itertools.product(range(size[0]), range(size[1]))


atoms_minimize = [(a, cp.Minimize) for a in atoms_minimize]
atoms_maximize = [(a, cp.Maximize) for a in atoms_maximize]


@pytest.mark.parametrize("atom_info, objective_type", atoms_minimize + atoms_maximize)
def test_constant_atoms(atom_info, objective_type) -> None:

    atom, size, args, obj_val = atom_info

    for indexer in get_indices(size):
        for solver in SOLVERS_TO_TRY:
            # Atoms with Constant arguments.
            prob_val = obj_val[indexer].value
            const_args = [Constant(arg) for arg in args]
            if len(size) != 0:
                objective = objective_type(atom(*const_args)[indexer])
            else:
                objective = objective_type(atom(*const_args))
            problem = Problem(objective)
            run_atom(atom, problem, prob_val, solver)

            # Atoms with Variable arguments.
            variables = []
            constraints = []
            for idx, expr in enumerate(args):
                variables.append(Variable(intf.shape(expr)))
                constraints.append(variables[-1] == expr)
            if len(size) != 0:
                objective = objective_type(atom(*variables)[indexer])
            else:
                objective = objective_type(atom(*variables))
            problem = Problem(objective, constraints)
            run_atom(atom, problem, prob_val, solver)

            # Atoms with Parameter arguments.
            parameters = []
            for expr in args:
                parameters.append(Parameter(intf.shape(expr)))
                parameters[-1].value = intf.DEFAULT_INTF.const_to_matrix(expr)
            if len(size) != 0:
                objective = objective_type(atom(*parameters)[indexer])
            else:
                objective = objective_type(atom(*parameters))
            run_atom(atom, Problem(objective), prob_val, solver)



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import warnings

import numpy as np
import pytest
import scipy.sparse as sp
from numpy.testing import assert_allclose, assert_equal
from scipy.linalg import ldl

import cvxpy as cp
from cvxpy.atoms.quad_form import decomp_quad
from cvxpy.settings import EIGVAL_TOL
from cvxpy.tests.base_test import BaseTest


class TestDecompQuad(BaseTest):
    """Tests for the decomp_quad helper function."""

    @staticmethod
    def _check(P) -> None:
        """Assert P = scale * (M1 @ M1.T - M2 @ M2.T) within tolerance."""
        scale, M1, M2 = decomp_quad(P)
        P_dense = P.toarray() if sp.issparse(P) else np.asarray(P)
        pos = scale * (M1 @ M1.conj().T) if M1.size else np.zeros_like(P_dense)
        neg = scale * (M2 @ M2.conj().T) if M2.size else np.zeros_like(P_dense)
        assert_allclose(P_dense, pos - neg, atol=1e-10)

    def test_psd_nsd(self) -> None:
        rng = np.random.default_rng(0)
        for n in (2, 5, 20):
            A = rng.standard_normal((n, n))
            self._check(A @ A.T)           # PSD, full rank
            self._check(-(A @ A.T))        # NSD
        A = rng.standard_normal((10, 7))
        self._check(A @ A.T)               # PSD, rank-deficient

    def test_indefinite(self) -> None:
        rng = np.random.default_rng(0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for n in (3, 10, 30):
                A = rng.standard_normal((n, n))
                P = A + A.T
                # Verify LDL uses nontrivial permutation
                _, _, perm = ldl(P)
                assert not np.array_equal(perm, np.arange(n))
                self._check(P)

    def test_2x2_blocks(self) -> None:
        """Zero diagonal forces Bunch-Kaufman 2x2 pivots."""
        rng = np.random.default_rng(0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for n in (6, 10, 30):
                A = rng.standard_normal((n, n))
                P = A + A.T
                np.fill_diagonal(P, 0.0)
                _, d, _ = ldl(P)
                assert np.any(np.diag(d, -1) != 0)
                self._check(P)

    def test_special(self) -> None:
        scale, M1, M2 = decomp_quad(np.zeros((4, 4)))
        assert_equal(scale, 0.0)
        assert_equal(M1.size, 0)
        assert_equal(M2.size, 0)
        self._check(np.diag([3.0, 0.0, 5.0, 0.0]))

    def test_complex_hermitian(self) -> None:
        rng = np.random.default_rng(0)
        A = rng.standard_normal((5, 5)) + 1j * rng.standard_normal((5, 5))
        self._check(A @ A.conj().T)        # PSD
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._check(A + A.conj().T)    # indefinite

    def test_sparse(self) -> None:
        self._check(sp.eye_array(5, format="csc") * 3.0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._check(sp.diags_array([1., -1, 2, -2, 3], format="csc"))

    def test_sparse_nsd(self) -> None:
        """Regression test: sparse NSD must use row permutation L[p,:], not column L[:,p].

        For NSD matrices, decomp_quad calls sparse_cholesky(-P) which returns (L, p) with
        L[p,:] @ L[p,:].T == -P.  The fix is M2 = L[p,:] (row perm), not L[:,p] (col perm).
        A tridiagonal matrix produces a non-trivial AMD ordering that exposes the bug.
        """
        n = 6
        off = np.ones(n - 1)
        # Negative discrete Laplacian: NSD, tridiagonal, AMD gives non-identity permutation.
        P_nsd = sp.diags([-4 * np.ones(n), off, off], [0, -1, 1], format="csc")
        self._check(P_nsd)


class TestNonOptimal(BaseTest):
    def test_singular_quad_form(self) -> None:
        """Test quad form with a singular matrix.
        """
        # Solve a quadratic program.
        np.random.seed(1234)
        for n in (3, 4, 5):
            for i in range(5):

                # construct a random 1d finite distribution
                v = np.exp(np.random.randn(n))
                v = v / np.sum(v)

                # construct a random positive definite matrix
                A = np.random.randn(n, n)
                Q = np.dot(A, A.T)

                # Project onto the orthogonal complement of v.
                # This turns Q into a singular matrix with a known nullspace.
                E = np.identity(n) - np.outer(v, v) / np.inner(v, v)
                Q = np.dot(E, np.dot(Q, E.T))
                observed_rank = np.linalg.matrix_rank(Q)
                desired_rank = n-1
                assert_equal(observed_rank, desired_rank)

                for action in 'minimize', 'maximize':

                    # Look for the extremum of the quadratic form
                    # under the simplex constraint.
                    x = cp.Variable(n)
                    if action == 'minimize':
                        q = cp.quad_form(x, Q)
                        objective = cp.Minimize(q)
                    elif action == 'maximize':
                        q = cp.quad_form(x, -Q)
                        objective = cp.Maximize(q)
                    constraints = [0 <= x, cp.sum(x) == 1]
                    p = cp.Problem(objective, constraints)
                    p.solve(solver=cp.OSQP)

                    # check that cvxpy found the right answer
                    xopt = x.value.flatten()
                    yopt = np.dot(xopt, np.dot(Q, xopt))
                    assert_allclose(yopt, 0, atol=1e-3)
                    assert_allclose(xopt, v, atol=1e-3)

    def test_sparse_quad_form(self) -> None:
        """Test quad form with a sparse matrix.
        """
        Q = sp.eye_array(2)
        x = cp.Variable(2)
        cost = cp.quad_form(x, Q)
        prob = cp.Problem(cp.Minimize(cost), [x == [1, 2]])
        self.assertAlmostEqual(prob.solve(solver=cp.OSQP), 5)

        # Here are our QP factors
        A = cp.Constant(sp.eye_array(4))
        c = np.ones(4).reshape((1, 4))

        # Here is our optimization variable
        x = cp.Variable(4)

        # And the QP problem setup
        function = cp.quad_form(x, A) - cp.matmul(c, x)
        objective = cp.Minimize(function)
        problem = cp.Problem(objective)

        problem.solve(solver=cp.OSQP)
        self.assertEqual(len(function.value), 1)

    def test_param_quad_form(self) -> None:
        """Test quad form with a parameter.
        """
        P = cp.Parameter((2, 2), PSD=True)
        Q = np.eye(2)
        x = cp.Variable(2)
        cost = cp.quad_form(x, P)
        P.value = Q
        prob = cp.Problem(cp.Minimize(cost), [x == [1, 2]])
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.assertAlmostEqual(prob.solve(solver=cp.SCS), 5)

    def test_non_symmetric(self) -> None:
        """Test when P is constant and not symmetric.
        """
        P = np.array([[2, 2], [3, 4]])
        x = cp.Variable(2)
        with self.assertRaises(Exception) as cm:
            cp.quad_form(x, P)
        self.assertTrue("Quadratic form matrices must be symmetric/Hermitian."
                        in str(cm.exception))

    def test_non_psd(self) -> None:
        """Test error when P is symmetric but not definite.
        """
        P = np.array([[1, 0], [0, -1]])
        x = cp.Variable(2)
        # Forming quad_form is okay
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cost = cp.quad_form(x, P)
        prob = cp.Problem(cp.Minimize(cost), [x == [1, 2]])
        with self.assertRaises(Exception) as cm:
            prob.solve(solver=cp.SCS)
        self.assertTrue("Problem does not follow DCP rules."
                        in str(cm.exception))

    def test_psd_exactly_tolerance(self) -> None:
        """Test that PSD check when eigenvalue is exactly -EIGVAL_TOL
        """
        P = np.array([[-0.999*EIGVAL_TOL, 0], [0, 10]])
        x = cp.Variable(2)
        # Forming quad_form is okay
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cost = cp.quad_form(x, P)
            prob = cp.Problem(cp.Minimize(cost), [x == [1, 2]])
            prob.solve(solver=cp.SCS)

    def test_nsd_exactly_tolerance(self) -> None:
        """Test that NSD check when eigenvalue is exactly EIGVAL_TOL
        """
        P = np.array([[0.999*EIGVAL_TOL, 0], [0, -10]])
        x = cp.Variable(2)
        # Forming quad_form is okay
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cost = cp.quad_form(x, P)
            prob = cp.Problem(cp.Maximize(cost), [x == [1, 2]])
            prob.solve(solver=cp.SCS)

    def test_obj_eval(self) -> None:
        """Test case where objective evaluation differs from result.
        """
        x = cp.Variable((2, 1))
        A = np.array([[1.0]])
        B = np.array([[1.0, 1.0]]).T
        obj0 = -B.T @ x
        obj1 = cp.quad_form(B.T @ x, A)
        prob = cp.Problem(cp.Minimize(obj0 + obj1))
        prob.solve(solver=cp.SCS)
        self.assertAlmostEqual(prob.value, prob.objective.value)

    def test_zero_term(self) -> None:
        """Test a quad form multiplied by zero.
        """
        data_norm = np.random.random(5)
        M = np.random.random(5*2).reshape((5, 2))
        c = cp.Variable(M.shape[1])
        lopt = 0
        laplacian_matrix = np.ones((2, 2))
        design_matrix = cp.Constant(M)
        objective = cp.Minimize(
            cp.sum_squares(design_matrix @ c - data_norm) +
            lopt * cp.quad_form(c, laplacian_matrix)
        )
        constraints = [(M[0] @ c) == 1]  # (K * c) >= -0.1]
        prob = cp.Problem(objective, constraints)
        prob.solve(solver=cp.SCS)

    def test_zero_matrix(self) -> None:
        """Test quad_form with P = 0.
        """
        x = cp.Variable(3)
        A = np.eye(3)
        b = np.ones(3,)
        c = -np.ones(3,)
        P = np.zeros((3, 3))
        expr = (1/2) * cp.quad_form(x, P) + c.T @ x
        prob = cp.Problem(cp.Minimize(expr),
                          [A @ x <= b])
        prob.solve(solver=cp.SCS)

    def test_assume_psd(self) -> None:
        """Test assume_PSD argument.
        """
        x = cp.Variable(3)
        A = np.eye(3)
        expr = cp.quad_form(x, A, assume_PSD=True)
        assert expr.is_convex()

        A = -np.eye(3)
        expr = cp.quad_form(x, A, assume_PSD=True)
        assert expr.is_convex()

        prob = cp.Problem(cp.Minimize(expr))
        # Transform to a SolverError.
        with pytest.raises(cp.SolverError):
            prob.solve(solver=cp.OSQP)

    def test_indefinite_assume_psd_raises(self) -> None:
        """quad_form_canon must raise ValueError for indefinite P with assume_PSD=True.

        In CVXPY 1.8.1 and earlier, the M1 (positive-eigenvalue) term was silently dropped.
        For P=diag(2,-1) and x=[1,2], the true x^T P x = -2, but the old
        code returned -4 with OPTIMAL status.
        """
        P = np.array([[2.0, 0.0], [0.0, -1.0]])  # indefinite: eigenvalues 2 and -1
        x = cp.Variable(2)
        expr = cp.quad_form(x, P, assume_PSD=True)
        prob = cp.Problem(cp.Minimize(expr), [x == [1.0, 2.0]])
        # use_quad_obj=False forces the conic canonicalization path where
        # quad_form_canon is called (vs. the QP path used by SCS >= 3.0).
        with pytest.raises(ValueError) as exc_info:
            prob.solve(solver=cp.SCS, use_quad_obj=False)
        assert "indefinite" in str(exc_info.value)
        assert "PSD" in str(exc_info.value)

    def test_quad_form_monotonicity_in_x(self) -> None:
        """Test DCP compositions requiring correct quad_form monotonicity in x."""
        P = np.array([[2.0, 0.5], [0.5, 1.0]])

        # x = square(z) is convex nonneg; quad_form nondecreasing in nonneg x → convex.
        z = cp.Variable(2, nonneg=True)
        assert cp.quad_form(cp.square(z), P).is_convex()

        # x = square(z) - 1 is convex but can be negative; monotonicity unknown → not DCP.
        assert not cp.quad_form(cp.square(z) - 1, P).is_convex()

    def test_quad_form_monotonicity_in_P(self) -> None:
        """Test DCP with P as a convex expression of a variable."""
        y = cp.Variable(2, nonneg=True)
        P = cp.diag(cp.square(y))

        # Nonneg x: quad_form is nondecreasing in P, so convex ∘ nondecreasing = convex.
        assert cp.quad_form(np.array([1.0, 2.0]), P).is_convex()

        # Mixed-sign x: monotonicity unknown, so not DCP-convex.
        assert not cp.quad_form(np.array([1.0, -2.0]), P).is_convex()



"""
Copyright 2022, the CVXPY Authors

Licensed under the Apache License, Version 2.0 (the "License");

you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

import cvxpy.interface as intf
import cvxpy.settings as s
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.qp_solvers.qp_solver import QpSolver
from cvxpy.utilities.citations import CITATION_DICT


class PROXQP(QpSolver):
    """QP interface for the PROXQP solver"""

    MIP_CAPABLE = False
    BOUNDED_VARIABLES = True

    # Map of Proxqp status to CVXPY status.
    STATUS_MAP = {"PROXQP_SOLVED": s.OPTIMAL,
                  "PROXQP_MAX_ITER_REACHED": s.USER_LIMIT,
                  "PROXQP_PRIMAL_INFEASIBLE": s.INFEASIBLE,
                  "PROXQP_DUAL_INFEASIBLE": s.UNBOUNDED}

    VAR_MAP = {"P": "H",
               "q": "g",
               "A": "A",
               "b": "b",
               "F": "C",
               "l": "lb",
               "G": "ub"}

    def name(self):
        return s.PROXQP

    def import_solver(self) -> None:
        import proxsuite
        proxsuite

    def invert(self, solution, inverse_data):
        attr = {s.SOLVE_TIME: solution.info.run_time}
        attr[s.EXTRA_STATS] = {"solution": solution}

        # Map PROXQP statuses back to CVXPY statuses
        status = self.STATUS_MAP.get(solution.info.status.name, s.SOLVER_ERROR)

        if status in s.SOLUTION_PRESENT:
            opt_val = solution.info.objValue + inverse_data[s.OFFSET]
            primal_vars = {
                PROXQP.VAR_ID:
                intf.DEFAULT_INTF.const_to_matrix(np.array(solution.x))
            }

            # Build dual vars dict keyed by constraint IDs
            # PROXQP returns solution.y (eq_duals) and solution.z (ineq_duals)
            eq_dual = utilities.get_dual_values(
                solution.y,
                utilities.extract_dual_value,
                inverse_data[self.EQ_CONSTR])
            ineq_dual = utilities.get_dual_values(
                solution.z,
                utilities.extract_dual_value,
                inverse_data[self.NEQ_CONSTR])
            dual_vars = {}
            dual_vars.update(eq_dual)
            dual_vars.update(ineq_dual)
            attr[s.NUM_ITERS] = solution.info.iter
            sol = Solution(status, opt_val, primal_vars, dual_vars, attr)
        else:
            sol = failure_solution(status, attr)
        return sol

    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts,
                       solver_cache=None):
        import proxsuite

        solver_opts['backend'] = solver_opts.get('backend', 'dense')
        backend = solver_opts['backend']

        if backend == "dense":
            # Convert sparse to dense matrices
            P = data[s.P].toarray()
            A = data[s.A].toarray()
            F = data[s.F].toarray()
        elif backend == "sparse":
            P = data[s.P]
            A = data[s.A]
            F = data[s.F]
        else:
            raise ValueError("Wrong input, backend most be either dense or sparse")

        q = data[s.Q]
        b = data[s.B]
        g = data[s.G]

        lb = -np.inf*np.ones(data[s.G].shape)
        data['lb'] = lb

        # Variable bounds
        x_lb = data.get(s.LOWER_BOUNDS)
        x_ub = data.get(s.UPPER_BOUNDS)

        n_var = data['n_var']
        n_eq = data['n_eq']
        n_ineq = data['n_ineq']

        # Overwrite default values
        solver_opts['eps_abs'] = solver_opts.get('eps_abs', 1e-8)
        solver_opts['eps_rel'] = solver_opts.get('eps_rel', 0.)
        solver_opts['max_iter'] = solver_opts.get('max_iter', 10000)
        solver_opts['rho'] = solver_opts.get('rho', 1e-6)
        solver_opts['mu_eq'] = solver_opts.get('mu_eq', 1e-3)
        solver_opts['mu_in'] = solver_opts.get('mu_in', 1e-1)
        # Use cached data
        if warm_start and solver_cache is not None and self.name() in solver_cache:
            solver, old_data, results = solver_cache[self.name()]
            new_args = {}
            for key in ['q', 'b', 'G', 'lb']:
                if any(data[key] != old_data[key]):
                    new_args[self.VAR_MAP[key]] = data[key]
            if P.data.shape != old_data[s.P].data.shape or any(
                    P.data != old_data[s.P].data):
                new_args['H'] = P
            if A.data.shape != old_data[s.A].data.shape or any(
                    A.data != old_data[s.A].data):
                new_args['A'] = A
            if F.data.shape != old_data[s.F].data.shape or any(
                    F.data != old_data[s.F].data):
                new_args['C'] = F

            old_x_lb = old_data.get(s.LOWER_BOUNDS)
            old_x_ub = old_data.get(s.UPPER_BOUNDS)
            if x_lb is not None and (old_x_lb is None or any(x_lb != old_x_lb)):
                new_args['l_box'] = x_lb
            if x_ub is not None and (old_x_ub is None or any(x_ub != old_x_ub)):
                new_args['u_box'] = x_ub

            if new_args:
                solver.update(**new_args)

            status = self.STATUS_MAP.get(results.info.status.name, s.SOLVER_ERROR)

            if status == s.OPTIMAL:
                solver.solve(results.x, results.y, results.z)
            else:
                solver.solve()
        else:
            has_box = x_lb is not None or x_ub is not None
            if backend == "dense":
                solver = proxsuite.proxqp.dense.QP(n_var, n_eq, n_ineq, has_box)
            elif backend == "sparse":
                solver = proxsuite.proxqp.sparse.QP(n_var, n_eq, n_ineq, has_box)

            init_kwargs = dict(
                        H=P,
                        g=q,
                        A=A,
                        b=b,
                        C=F,
                        l=lb,
                        u=g,
                        rho=solver_opts['rho'],
                        mu_eq=solver_opts['mu_eq'],
                        mu_in=solver_opts['mu_in'])
            if x_lb is not None:
                init_kwargs['l_box'] = x_lb
            if x_ub is not None:
                init_kwargs['u_box'] = x_ub
            solver.init(**init_kwargs)

            solver.settings.eps_abs = solver_opts['eps_abs']
            solver.settings.eps_rel = solver_opts['eps_rel']
            solver.settings.max_iter = solver_opts['max_iter']
            solver.settings.verbose = verbose

            solver.solve()

        results = solver.results

        if solver_cache is not None:
            solver_cache[self.name()] = (solver, data, results)

        return results

    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT["PROXQP"]